<?php

namespace App\Http\Controllers;

use App\Models\User;
use App\Models\Truck;
use App\Models\Driver;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Hash;
use Illuminate\Validation\Rule;


class DriverController extends Controller
{
    protected $created_by;

    /**
     * Create a new controller instance.
     *
     * @return void
     */
    public function __construct()
    {
        $this->middleware(function ($request, $next) {
            $this->created_by = auth()->user()->id;
            return $next($request);
        });
    }
    /**
     * Display a listing of the resource.
     */
    public function index()
    {
        $users = User::role('driver')->whereHas('driver', function ($query) {
            $query->where('created_by', $this->created_by);
        })->latest()->get();
        return view('driver.index', compact('users'));
    }

    /**
     * Show the form for creating a new resource.
     */
    public function create()
    {
        // $trucks = Truck::all();
        // $trailers = Trailer::all();
        return view('driver.create');
    }

    /**
     * Store a newly created resource in storage.
     */
    public function store(Request $request)
    {
        $request->validate([
            'name' => 'required',
            'email' => 'required',
            'phone' => 'required',
            'dob' => 'required',
            'address' => 'required',
            'city' => 'required',
            'state' => 'required',
            'zipcode' => 'required',
            'status' => 'required',
            'application_date' => 'required',
            'password' => ['required', 'string', 'min:8', 'confirmed'],
            // 'image' => 'required',
            // 'pay_to' => 'nullable', // This field can be nullable
            // 'co_driver' => 'nullable', // This field can be nullable
        ]);
        if ($request->hasFile('avatar')) {
            $avatar = storeFile(request('avatar'), 'storage/driver/');
        } else {
            $avatar ='avatar.png';
        }
        $user = new User();
        $user->name = $request->name;
        $user->email = $request->email;
        $user->password = Hash::make($request->password);
        $user->avatar = $avatar;
        $user->save();
        $user->assignRole('driver');

        $driver = new Driver();
        $driver->user_id = $user->id;
        $driver->name =  $request->name;
        $driver->email =  $request->email;
        $driver->phone =  $request->phone;
        $driver->address =  $request->address;
        $driver->dob =  $request->dob;
        $driver->city =  $request->city;
        $driver->state =  $request->state;
        $driver->zipcode =  $request->zipcode;
        $driver->status =  $request->status;
        $driver->application_date =  $request->application_date;
        $driver->save();
        return redirect()->route('drivers.index')->with('success', 'Driver Created Successfully');
        
    }

    /**
     * Display the specified resource.
     */
    public function show(string $id)
    {
        //
    }

    /**
     * Show the form for editing the specified resource.
     */
    public function edit(string $id)
    {
        $check = Driver::where('user_id', $id)->where('created_by', $this->created_by)->first();
        if ($check) {
            $user = User::findOrFail($id);
            return view('driver.create', compact('user'));
        } else {
            return redirect()->route('drivers.index')->with('error', 'Record not found');
        }
    }

    /**
     * Update the specified resource in storage.
     */
    public function update(Request $request, string $id)
    {
        $user = User::findOrFail($id);
        $request->validate([
            'name' => 'required',
            'email' => ['required', 'string', 'max:255', 'email', Rule::unique('users')->ignore($user->id)],
            'phone' => 'required',
            'dob' => 'required',
            'address' => 'required',
            'city' => 'required',
            'state' => 'required',
            'zipcode' => 'required',
            'status' => 'required',
            'application_date' => 'required',
        ]);
        if ($request->hasFile('avatar')) {
            $avatar = storeFile(request('avatar'), 'storage/driver/');
        } else {
            $avatar =  'avatar.png';
        }
        if ($request->password) {
            $request->validate([
                'password' => ['string', 'min:8', 'confirmed'],
            ]);
            $password = Hash::make($request->password);
        } else {
            unset($request->password);
        }
        $user->name = $request->name;
        $user->email = $request->email;
        if (isset($request->password)) {
            $user->password = $password;
        }
        $user->avatar = $avatar;
        $user->assignRole('driver');
        $user->update();
        $driver = Driver::where('user_id', $user->id)->first();
        $driver->name =  $request->name;
        $driver->email =  $request->email;
        $driver->phone =  $request->phone;
        $driver->address =  $request->address;
        $driver->dob =  $request->dob;
        $driver->city =  $request->city;
        $driver->state =  $request->state;
        $driver->zipcode =  $request->zipcode;
        $driver->status =  $request->status;
        $driver->application_date =  $request->application_date;
        $driver->update();
        return redirect()->route('drivers.index')->with('success', 'Driver Updated Successfully');
    }

    /**
     * Remove the specified resource from storage.
     */
    public function destroy(string $id)
    {
        $check = Driver::where('user_id', $id)->where('created_by', $this->created_by)->first();
        if ($check) {
            $user = User::findOrFail($id);
            $user->delete();
            return redirect()->route('drivers.index')->with('success', 'Driver deleted successfully.');
        } else {
            return redirect()->route('drivers.index')->with('error', 'Record not found');
        }

    }
}
